------------------------------------------------------------------------------
--  UTIA AV CR, v.v.i
--  Signal Processing Department
--  Pod Vodarenskou vezi 4
--  182 08 Prague
--  Czech Republic
--  web: http://zs.utia.cas.cz
--  contact: sp@utia.cz
--
--  Filename: im_hdmi_in.vhd
--  Date: 14.09.2016
--  Author: Lukas Kohout <kohoutl@utia.cas.cz>
--  Description:
--     HDMI in for AVNET FMC board
--
--  Changelog:
--
--
--
--
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

------------------------------------------------------------------------------
entity im_hdmi_in is
    generic (
        C_DATA_WIDTH  : natural := 16
    );
    port (
        -- 16-bit HDMI IN (YCrCB 4:2:2)
        hdmi_clk      : in  std_logic;
        hdmi_data     : in  std_logic_vector(C_DATA_WIDTH-1 downto 0);

        -- output vide data + syncs
        vid_data      : out  std_logic_vector(C_DATA_WIDTH-1 downto 0);
        vid_de        : out  std_logic;
        vid_vblank    : out  std_logic;
        vid_hblank    : out  std_logic
    );
end entity im_hdmi_in;

------------------------------------------------------------------------------
architecture rtl of im_hdmi_in is

------------------------------------------------------------------------------
    signal s_hdmi_data_ff         : std_logic_vector(C_DATA_WIDTH-1 downto 0);
    attribute IOB : string;
    attribute IOB of s_hdmi_data_ff: signal is "FORCE";


    constant C_DELAY_IN           : natural := 5;
    type data_pipe_t is array (C_DELAY_IN-1 downto 0) of std_logic_vector(C_DATA_WIDTH-1 downto 0);

    signal s_vid_in_ff            : data_pipe_t;

    -- start of active video
    constant C_SYNC_CODE_SAV_AV   : std_logic_vector(4*C_DATA_WIDTH-1 downto 0) := x"FFFF000000008080";
    -- end of active video
    constant C_SYNC_CODE_EAV_AV   : std_logic_vector(4*C_DATA_WIDTH-1 downto 0) := x"FFFF000000009D9D";

    -- start of blank
    constant C_SYNC_CODE_SAV_BL   : std_logic_vector(4*C_DATA_WIDTH-1 downto 0) := x"FFFF00000000ABAB";
    -- end of blank
    constant C_SYNC_CODE_EAV_BL   : std_logic_vector(4*C_DATA_WIDTH-1 downto 0) := x"FFFF00000000B6B6";

    signal s_sav_av               : std_logic;
    signal s_eav_av               : std_logic;
    signal s_sav_bl               : std_logic;
    signal s_eav_bl               : std_logic;

    constant C_DELAY_SYNC         : natural := 4;
    signal s_sav_av_ff            : std_logic_vector(C_DELAY_SYNC-1 downto 0);
    signal s_sav_bl_ff            : std_logic_vector(C_DELAY_SYNC-1 downto 0);


    signal s_vblank               : std_logic;
    signal s_hblank               : std_logic;
    signal s_de                   : std_logic;


------------------------------------------------------------------------------
begin

------------------------------------------------------------------------------
-- registers, delay lines
    p_ffs : process (hdmi_clk) is
    begin
        if hdmi_clk'event and hdmi_clk = '1' then

            -- force IOB
            s_hdmi_data_ff <= hdmi_data;

            -- make a pipe from the input data to get sync codes
            s_vid_in_ff <= s_vid_in_ff(C_DELAY_IN-2 downto 0) & s_hdmi_data_ff; --hdmi_data;


            -- delay detected sync codes to generate separaterd sync synchronously with the data
            s_sav_av_ff <= s_sav_av_ff(C_DELAY_SYNC-2 downto 0) & s_sav_av;
            s_sav_bl_ff <= s_sav_bl_ff(C_DELAY_SYNC-2 downto 0) & s_sav_bl;

        end if;
    end process p_ffs;



------------------------------------------------------------------------------
-- decode HDMI data - separate video data and syncs according to ccir656
    p_decode : process ( s_vid_in_ff ) is
    variable v_sync_code : std_logic_vector(4*C_DATA_WIDTH-1 downto 0);
    begin

        v_sync_code :=  s_vid_in_ff(3) & s_vid_in_ff(2) & s_vid_in_ff(1) & s_vid_in_ff(0);

        -- start of active video
        if v_sync_code = C_SYNC_CODE_SAV_AV then
            s_sav_av  <= '1';
        else
            s_sav_av  <= '0';
        end if;

        -- end of active video
        if v_sync_code = C_SYNC_CODE_EAV_AV then
            s_eav_av  <= '1';
        else
            s_eav_av  <= '0';
        end if;

        -- start of video blank
        if v_sync_code = C_SYNC_CODE_SAV_BL then
            s_sav_bl  <= '1';
        else
            s_sav_bl  <= '0';
        end if;

        -- end of video blank
        if v_sync_code = C_SYNC_CODE_EAV_BL then
            s_eav_bl  <= '1';
        else
            s_eav_bl  <= '0';
        end if;

    end process p_decode;


------------------------------------------------------------------------------
-- generarte - separated synchronization pulses for output video data
    p_out_sync_gen : process (hdmi_clk) is
    begin
        if hdmi_clk'event and hdmi_clk = '1' then

            -- generate active video signal vid_de
            if  s_sav_av_ff(3) = '1' then
                s_de  <= '1';
            elsif (s_eav_av or s_eav_bl) = '1' then
                s_de  <= '0';
            end if;

            -- generate horizontal blank signal vid_hblank
            if (s_eav_av or s_eav_bl) = '1' then
                s_hblank  <= '1';
            elsif (s_sav_av_ff(3) or s_sav_bl_ff(3)) = '1' then
                s_hblank  <= '0';
            end if;

            -- generate vertical blank signal vid_vblank
            if (s_sav_bl or s_eav_bl) = '1' then
                s_vblank  <= '1';
            elsif s_sav_av_ff(3) = '1' then
                s_vblank  <= '0';
            end if;

      end if;
   end process p_out_sync_gen;


------------------------------------------------------------------------------
-- outpu FFs
    p_ffs_out : process (hdmi_clk) is
    begin
        if hdmi_clk'event and hdmi_clk = '1' then
            vid_vblank <= s_vblank;
            vid_hblank <= s_hblank;
            vid_de     <= s_de;
            vid_data   <= s_vid_in_ff(4);
        end if;
    end process p_ffs_out;


end rtl;
